/*
 * Copyright (c) 2005, University of Massachusetts
 * All Rights Reserved.
 * 
 * Name:           jacobi_support.c
 * Last modified:  30 Mar 2005
 * Author:         Stephen Siegel <siegel@cs.umass.edu>
 *
 * C support code for model of the Jacobi iteration program from
 * Karniadakis & Kirby.
 */

#include "../symcode/seqpar.c"

int widths[N][N];
int column_widths[N];

extern void calculate_widths(struct BigMatrix, int, int);
extern void prettyprint_vector(struct LongVector, int);
extern void prettyprint_matrix(struct BigMatrix, int, int);
extern void printParallelummary();
extern void report_par_error(struct LongVector, int);
extern void print_cond_warning(int, int);


void initializeSystem() {
  int i,j;

  init_seqpar();
  printf("Beginning verification of Jacobi program to solve Ax=b.\n\n");
  printf("    NPROCS = %d\n", NPROCS);
  printf("         N = %d\n", N);
  printf("     MAXIT = %d\n", MAXIT);
  printf("\n");
  for (i=0; i<N; i++) {
    for (j=0; j<N; j++) {
      now.A_global.row[i].entryAt[j] = SYM_symbolicConstant_0(i*N+j);
    }
  }
  for (i=0; i<N; i++) {
    now.b_global.entryAt[i] = SYM_symbolicConstant_0(N*N+i);
  }
  SYM_symbolicConstant_0(N*N+N); /* ABSTOL */
  printf("Input Matrix A:\n\n");
  prettyprint_matrix(now.A_global, N, N);
  printf("\n");
  printf("Input Vector b:\n\n");
  prettyprint_vector(now.b_global, N);
  printf("\n");
  printf("Tolerance (abstol): ");
  SYM_printExpr(ABSTOL);
  printf("\n\n");
  printf("-------------------------------");
  printf("--------------------------------------\n\n");
}


void printSolution() {
  update_seqpar();
#if VERBOSE>0
  printf("\nResult vector x:\n\n");
  prettyprint_vector(now.x_global, N);
#endif
}


void calculate_widths(struct BigMatrix matrix, int num_rows, int num_cols) {
  int col, row, expr, width;

  for (col = 0; col < num_cols; col++) {
    column_widths[col] = 0;
    for (row = 0; row < num_rows; row++) {
      expr = matrix.row[row].entryAt[col];
      width = SYM_exprLength(expr);
      widths[row][col] = width;
      if (width>column_widths[col]) {
        column_widths[col] = width;
      }
    }
    column_widths[col]++;
  }
}


void prettyprint_vector(struct LongVector vec, int num_rows) {
  int row;

  for (row = 0; row < num_rows; row++) {
    SYM_printExpr(vec.entryAt[row]);
    printf("\n");
  }
  printf("\n");
}


void prettyprint_matrix(struct BigMatrix matrix, int num_rows, int num_cols) {
  int row, col, spaces, i;

  calculate_widths(matrix, num_rows, num_cols);
  for (row = 0; row < num_rows; row++) {
    for (col = 0; col < num_cols; col++) {
      SYM_printExpr(matrix.row[row].entryAt[col]);
      if (col < num_cols-1) {
	spaces = column_widths[col]-widths[row][col];
	for (i = 0; i < spaces; i++) printf(" ");
      }
    }
    printf("\n");
  }
  printf("\n");
}



void report_par_error(struct LongVector vec, int pos) {
  incr_error();
  printf("\nSequential and parralel solutions differ in position %d:\n\n",
         pos);
  SYM_printExpr(now.x_global.entryAt[pos]);
  printf("\n\n");
  SYM_printExpr(vec.entryAt[pos]);
  printf("\n");
}


void print_cond_warning(int expr, int rhs) {
  printf("WARNING: unable to determine truth value of boolean expression\n");
  printf("occurring in parallel program...exploring both possibilities:\n\n");
  SYM_printExpr(expr);
  printf(" < ");
  SYM_printExpr(rhs);
  printf("\n\n");
}
