/*
 * Copyright (c) 2005, University of Massachusetts
 * All Rights Reserved.
 * 
 * Name:           sym.c 
 * Last modified:  30 Mar 2005
 * Author:         Stephen Siegel <siegel@cs.umass.edu>
 *
 * This is a library of functions for manipulating symbolic
 * expressions and maintaining the symbolic expression table.
 */


/* Default values for constants */

/* The max nesting depth of an expression */
#ifndef SYM_MAX_DEPTH
#define SYM_MAX_DEPTH 500
#endif

/* The max number of expressions */
#ifndef SYM_MAX_EXPR
#define SYM_MAX_EXPR 10000
#endif

/* The length of the expression hashtable */
#ifndef SYM_HASHTABLE_LEN
#define SYM_HASHTABLE_LEN 8192
#endif

/* Operation codes */

#define SYM_PLUS        1
#define SYM_MINUS       2
#define SYM_TIMES       3
#define SYM_DIVIDE      4
#define SYM_NEG         5
#define SYM_SQRT        6
#define SYM_ABS         7
#define SYM_EQUALS      8
#define SYM_NEQUALS     9
#define SYM_LT          10
#define SYM_GT          11
#define SYM_LTE         12
#define SYM_GTE         13
#define SYM_BLT         14 /* BLT(x,y)=1 if x<y, else = 0 */
#define SYM_BGTE        15 /* BGTE(x,y)=1 if x>=y else = 0 */
#define SYM_INT         16 /* integer constant n */
#define SYM_RATIONAL    17 /* rational constant p/q */
#define SYM_SYM_CONST   18 /* symbolic constant x_n */
#define SYM_AND         19


/* Special Constants */

#define SYM_ZERO          0
#define SYM_ONE           1
#define SYM_FALSE         SYM_ZERO
#define SYM_TRUE          SYM_ONE
#define SYM_HERBRAND      0
#define SYM_IEEE754       1
#define SYM_REAL          2
#define MAX(x,y) ((x>y) ? x : y) 


/* Types */

struct Sym_expr {
  int operator;            /* arithmetic operator code */
  struct Sym_expr *left;   /* left operand */
  struct Sym_expr *right;  /* right operand */
  int val;                 /* for leaf expressions */
  int id;                  /* unique id */
  struct Sym_expr *next;   /* next elt with same hash */
};

typedef struct Sym_expr Sym_Expr;


typedef struct Sym_sum_list {
  int       length;              /* number of non-integral terms */
  int       intterm;             /* integral term (possibly 0)   */
  Sym_Expr* list[SYM_MAX_DEPTH]; /* list of non-integral terms   */
} Sym_SumList;


/* Static Global Variables */

Sym_Expr* Sym_exprArray[SYM_MAX_EXPR];      /* array by id of exprs */
Sym_Expr* Sym_hashtable[SYM_HASHTABLE_LEN]; /* array by hash of exprs */
int Sym_nExpr=0;                            /* current number of exprs */
int Sym_equivalenceType = SYM_HERBRAND;
Sym_Expr* Sym_zero;
Sym_Expr* Sym_one;
Sym_Expr* Sym_true;
Sym_Expr* Sym_false;
Sym_SumList* Sym_sumlist1;
Sym_SumList* Sym_sumlist2;
Sym_SumList* Sym_sumlist3;

/* Function Prototypes */

void Sym_printSymTable();


/* Function Definitions: private */

void Sym_assert(int a, char* s) {
  if (!a) {
    printf("Assertion violated: %s", s);
    Sym_printSymTable();
    exit(1);
  }
}


Sym_Expr* Sym_getExpr(int id) {
  Sym_assert(id<Sym_nExpr, "getExpr: illegal id");
  return Sym_exprArray[id];
}


int Sym_hashFunc(int operator, int leftid, int rightid, int val) {
  return (operator+16*leftid+256*(rightid+val))%SYM_HASHTABLE_LEN;
}


int Sym_hashFor(int operator, Sym_Expr* left, Sym_Expr* right, int val) {
  int leftid,rightid;

  if (left) leftid=left->id; else leftid=0;
  if (right) rightid=right->id; else rightid=0;
  return Sym_hashFunc(operator,leftid,rightid,val);
}

int Sym_hash(Sym_Expr* expr) {
  Sym_assert(expr!=0, "null expr in SYM_hash");

  return Sym_hashFor(expr->operator, expr->left, expr->right, expr->val);
}


/*
 * Returns the length of the string s, and prints it out too, if
 * visible != 0.
 */
int Sym_outputString(char* s, int visible) {
  int j=0;

  while (s[j] != '\0') {
    j++;
  }
  if (visible) {
    printf("%s", s);
  }
  return j;
}


int Sym_intLength(int n) {
  if (n < 0) {
    return 1+Sym_intLength(-n);
  }
  else {
    int result = 1;
    
    while (n >= 10) {
      n=n/10;
      result++;
    }
    return result;
  }
}


/*
 * Returns the string length of the integer n, and prints it out too,
 * if visible != 0.
 */
int Sym_outputInt(int n, int visible) {
  if (visible) {
    printf("%d", n);
  }
  return Sym_intLength(n);
}


int Sym_outputExpr(Sym_Expr* expr, int visible) {
  int operator, val, leftop, rightop, result;
  Sym_Expr* left;
  Sym_Expr* right;

  Sym_assert(expr!=0, "outputExpr: null expr");
  operator = expr->operator;
  left = expr->left;
  right = expr->right;
  val = expr->val;
  result = 0;
  if (expr==Sym_zero) {
    result += Sym_outputString("0", visible);
  }
  else if (expr == Sym_one) {
    result += Sym_outputString("1", visible);
  }  
  else if (operator == SYM_PLUS) {
    result += 
      Sym_outputExpr(left, visible) +
      Sym_outputString("+", visible) +
      Sym_outputExpr(right, visible);
  }
  else if (operator == SYM_MINUS) {
    result +=
      Sym_outputExpr(left, visible) +
      Sym_outputString("-", visible);
    rightop = right->operator;
    if (rightop == SYM_PLUS || rightop == SYM_MINUS) {
      result +=
	Sym_outputString("(", visible) +
	Sym_outputExpr(right, visible) +
	Sym_outputString(")", visible);
    }
    else {
      result += Sym_outputExpr(right, visible);
    }
  }
  else if (operator == SYM_TIMES) {
    leftop = left->operator;
    rightop = right->operator;
    if (leftop == SYM_PLUS || leftop == SYM_MINUS || leftop == SYM_DIVIDE) {
      result +=
	Sym_outputString("(", visible) +
	Sym_outputExpr(left, visible) +
	Sym_outputString(")", visible);
    }
    else {
      result += Sym_outputExpr(left, visible);
    }
    result += Sym_outputString("*", visible);
    if (rightop == SYM_PLUS || rightop == SYM_MINUS ||
	rightop == SYM_DIVIDE) {
      result +=
	Sym_outputString("(", visible) +
	Sym_outputExpr(right, visible) +
	Sym_outputString(")", visible);
    }
    else {
      result += Sym_outputExpr(right, visible);
    }
  }
  else if (operator == SYM_DIVIDE) {
    leftop = left->operator;
    rightop = right->operator;
    if (leftop == SYM_PLUS || leftop == SYM_MINUS ||
        leftop == SYM_TIMES || leftop == SYM_DIVIDE) {
      result +=
	Sym_outputString("(", visible) +
	Sym_outputExpr(left, visible) +
	Sym_outputString(")", visible);
    }
    else {
      result += Sym_outputExpr(left, visible);
    }
    result += Sym_outputString("/", visible);
    if (rightop == SYM_PLUS || rightop == SYM_MINUS ||
        rightop == SYM_TIMES || rightop == SYM_DIVIDE) {
      result +=
	Sym_outputString("(", visible) +
	Sym_outputExpr(right, visible) +
	Sym_outputString(")", visible);
    }
    else {
      result += Sym_outputExpr(right, visible);
    }
  }
  else if (operator == SYM_NEG) {
    result +=  Sym_outputString("--", visible);
    leftop = left->operator;
    if (leftop == SYM_PLUS || leftop == SYM_MINUS) {
      result +=
	Sym_outputString("(", visible) +
	Sym_outputExpr(left, visible) +
	Sym_outputString(")", visible);
    }
    else {
      result += Sym_outputExpr(left, visible);
    }
  }
  else if (operator == SYM_SQRT) {
    result +=
      Sym_outputString("SQRT(", visible) +
      Sym_outputExpr(left, visible) +
      Sym_outputString(")", visible);
  }
  else if (operator == SYM_ABS) {
    result +=
      Sym_outputString("|", visible) +
      Sym_outputExpr(left, visible) +
      Sym_outputString("|", visible);
  }
  else if (operator == SYM_SYM_CONST) {
    result +=
      Sym_outputString("x", visible)+
      Sym_outputInt(val, visible);
  }
  else if (operator == SYM_INT) {
    result += Sym_outputInt(val, visible);
  }
  else if (operator == SYM_EQUALS) {
    result +=
      Sym_outputExpr(left, visible) +
      Sym_outputString(" == ", visible) +
      Sym_outputExpr(right, visible);
  }
  else if (operator == SYM_NEQUALS) {
    result += 
      Sym_outputExpr(left, visible) +
      Sym_outputString(" != ", visible) +
      Sym_outputExpr(right, visible);
  }
  else if (operator == SYM_LT) {
    result +=
      Sym_outputExpr(left, visible) +
      Sym_outputString(" < ", visible) +
      Sym_outputExpr(right, visible);
  }
  else if (operator == SYM_GT) {
    result +=
      Sym_outputExpr(left, visible) +
      Sym_outputString(" > ", visible) +
      Sym_outputExpr(right, visible);
  }
  else if (operator == SYM_LTE) {
    result +=
      Sym_outputExpr(left, visible) +
      Sym_outputString(" <= ", visible) +
      Sym_outputExpr(right, visible);
  }
  else if (operator == SYM_GTE) {
    result +=
      Sym_outputExpr(left, visible) +
      Sym_outputString(" >= ", visible) +
      Sym_outputExpr(right, visible);
  }
  else if (operator == SYM_BLT) {
    result +=
      Sym_outputString("BLT(", visible) +
      Sym_outputExpr(left, visible) +
      Sym_outputString(", ", visible) +
      Sym_outputExpr(right, visible) +
      Sym_outputString(")", visible);
  }
  else if (operator == SYM_BGTE) {
    result +=
      Sym_outputString("BGTE(", visible) +
      Sym_outputExpr(left, visible) +
      Sym_outputString(", ", visible) +
      Sym_outputExpr(right, visible) +
      Sym_outputString(")", visible);
  }
  else if (operator == SYM_AND) {
    result +=
      Sym_outputString("AND(", visible) +
      Sym_outputExpr(left, visible) +
      Sym_outputString(", ", visible) +
      Sym_outputExpr(right, visible) +
      Sym_outputString(")", visible);
  }
  else {
    printf("\nUnknown operator: %d\n", operator);
    result = -1;
  }
  return result;
}


int Sym_printExpr(Sym_Expr* expr) {
  return Sym_outputExpr(expr, 1);
}


int Sym_exprLength(Sym_Expr* expr) {
  return Sym_outputExpr(expr, 0);
}


void Sym_printExprShort(Sym_Expr* expr) {
  int operator, leftid, rightid, val;

  Sym_assert(expr!=0, "printExprShort: null expr");
  operator = expr->operator;
  val = expr->val;
  if (expr->left) leftid = expr->left->id; else leftid=0;
  if (expr->right) rightid = expr->right->id; else rightid=0;

  printf("(");
  switch (operator) {
  case SYM_PLUS: printf("+ e%d e%d", leftid, rightid); break;
  case SYM_MINUS: printf("- e%d e%d", leftid, rightid); break;
  case SYM_TIMES: printf("* e%d e%d", leftid, rightid); break;
  case SYM_DIVIDE: printf("/ e%d e%d", leftid, rightid); break;
  case SYM_AND: printf("AND e%d e%d", leftid, rightid); break;
  case SYM_NEG: printf("-- e%d", leftid); break;
  case SYM_SQRT: printf("SQRT e%d", leftid); break;
  case SYM_ABS: printf("ABS e%d", leftid); break;
  case SYM_EQUALS: printf("== e%d e%d", leftid, rightid); break;
  case SYM_NEQUALS: printf("!= e%d e%d", leftid, rightid); break;
  case SYM_LT: printf("< e%d e%d", leftid, rightid); break;
  case SYM_GT: printf("> e%d e%d", leftid, rightid); break;
  case SYM_LTE: printf("<= e%d e%d", leftid, rightid); break;
  case SYM_GTE: printf(">= e%d e%d", leftid, rightid); break;
  case SYM_BLT: printf("BLT e%d e%d", leftid, rightid); break;
  case SYM_BGTE: printf("BGTE e%d e%d", leftid, rightid); break;
  case SYM_INT: printf("INT %d", val); break;
  case SYM_SYM_CONST: printf("x %d", val); break;
  }
  printf(")");
}


void Sym_printSymTable() {
  int i;

  printf("\n\nSymbol Table:\n\n");
  for (i=0; i<Sym_nExpr; i++) {
    Sym_Expr* expr = Sym_getExpr(i);

    printf("e%d.  ", i);
    Sym_printExprShort(expr);
    printf("\n  ");
    Sym_printExpr(expr);
    printf("\n");
  }
  printf("\n");
}


void Sym_printSymTableShort() {
  int i;

  printf("\n\nSymbol Table:\n\n");
  for (i=0; i<Sym_nExpr; i++) {
    Sym_Expr* expr = Sym_getExpr(i);

    printf("e%d.  ", i);
    Sym_printExprShort(expr);
    printf("\n");
  }
  printf("\n");
}


/* Note: this assumes the path condition pc is an expression of the
 * following form (AND (AND (AND e1) e2) e3), etc.  With a little bit
 * of work, we could make this more general, but it is fine for now.
 */
void Sym_printPathCondition(Sym_Expr* pc) {
  int done, count;
  Sym_Expr* expr;

  if (pc == Sym_true) {
    printf("Path condition: true\n\n");
    return;
  }
  printf("Path condition is conjunction of the following:\n\n");
  done = 0;
  count = 0;
  while (!done) {
    if (pc->operator == SYM_AND) {
      expr = pc->right;
      pc = pc->left;
    }
    else {
      expr = pc;
      done = 1;
    }
    if (expr == Sym_true) {
    }
    else {    
      printf("  %d. ", count);
      Sym_printExpr(expr);
      printf("\n");
      count++;
    }
  }
}


/* Returns ptr to the matching Expr, or null if one does not exist */
Sym_Expr* Sym_findExpr(int operator, Sym_Expr* left,
		       Sym_Expr* right, int val) {
  Sym_Expr* expr;
  int hash = Sym_hashFor(operator, left, right, val);

  for (expr=Sym_hashtable[hash]; expr; expr=expr->next) {
    if (expr->operator==operator && expr->left==left &&
	expr->right==right && expr->val==val) {
      return expr;
    }
  }
  return 0; /* can't find */
}

Sym_Expr* Sym_findBinaryExpr(int operator, Sym_Expr* left, Sym_Expr* right) {
  return Sym_findExpr(operator,left,right,0);
}

Sym_Expr* Sym_findUnaryExpr(int operator, Sym_Expr* left) {
  return Sym_findExpr(operator,left,0,0);
}

Sym_Expr* Sym_findLeafExpr(int operator, int val) {
  return Sym_findExpr(operator,0,0,val);
}


Sym_Expr* Sym_addExpr(int operator, Sym_Expr* left, Sym_Expr* right, int val) {
  int hash = Sym_hashFor(operator, left, right, val);
  Sym_Expr* old = Sym_hashtable[hash];
  Sym_Expr* new = (Sym_Expr*)malloc(sizeof(Sym_Expr));

  Sym_assert(new != 0, "Symbolic package out of memory!");
  Sym_assert(Sym_nExpr<SYM_MAX_EXPR, "Too many expressions!");
  new->operator=operator;
  new->left=left;
  new->right=right;
  new->val=val;
  new->id=Sym_nExpr;
  new->next=0;
  Sym_exprArray[Sym_nExpr]=new;
  if (!old) {
    Sym_hashtable[hash]=new;
  }
  else {
    while (old->next) {
      old=old->next;
    }
    old->next=new;
  }
  Sym_nExpr++;
  return new;
}


Sym_Expr* Sym_addBinary(int operator, Sym_Expr* left, Sym_Expr* right) {
  return Sym_addExpr(operator, left, right, 0);
}


Sym_Expr* Sym_addUnary(int operator, Sym_Expr* left) {
  return Sym_addExpr(operator, left, 0, 0);
}


Sym_Expr* Sym_addLeaf(int operator, int val) {
  return Sym_addExpr(operator,0,0,val);
}


Sym_Expr* Sym_findOrAddExpr(int operator, Sym_Expr* left,
			    Sym_Expr* right, int val) {
  Sym_Expr* expr;
  Sym_Expr* prev;
  int hash = Sym_hashFor(operator, left, right, val);

  prev = 0;
  for (expr=Sym_hashtable[hash]; expr; expr=expr->next) {
    if (expr->operator==operator && expr->left==left &&
	expr->right==right && expr->val==val) {
      return expr;
    }
    prev=expr;
  }
  expr = (Sym_Expr*)malloc(sizeof(Sym_Expr));
  Sym_assert(expr != 0, "Symbolic package out of memory!");
  Sym_assert(Sym_nExpr<SYM_MAX_EXPR, "Too many expressions!");
  expr->operator=operator;
  expr->left=left;
  expr->right=right;
  expr->val=val;
  expr->id=Sym_nExpr;
  expr->next=0;
  Sym_exprArray[Sym_nExpr]=expr;
  if (!prev) {
    Sym_hashtable[hash] = expr;
  }
  else {
    prev->next = expr;
  }
  Sym_nExpr++;
  return expr;
}


Sym_Expr* Sym_findOrAddBinary(int operator, Sym_Expr* left, Sym_Expr* right) {
  return Sym_findOrAddExpr(operator,left,right,0);
}

Sym_Expr* Sym_findOrAddUnary(int operator, Sym_Expr* left) {
  return Sym_findOrAddExpr(operator, left, 0, 0);
}

Sym_Expr* Sym_findOrAddLeaf(int operator, int val) {
  return Sym_findOrAddExpr(operator, 0, 0, val);
}


int Sym_isInt(Sym_Expr* expr) {
  if (expr->operator==SYM_INT) {
    return 1;
  }
  if (expr->operator==SYM_NEG) {
    Sym_Expr* expr2 = expr->left;

    if (expr2->operator==SYM_INT) {
      return 1;
    }
    else {
      return 0;
    }
  }
  return 0;
}


int Sym_intVal(Sym_Expr* expr) {
  if (expr->operator==SYM_INT) {
    return expr->val;
  }
  if (expr->operator==SYM_NEG) {
    Sym_Expr* expr2 = expr->left;

    if (expr2->operator==SYM_INT) {
      return -(expr2->val);
    }
    else {
      Sym_assert(0, "not int expr");
    }
  }
  return 0;
}


/* Level 0 arithmetic */

/* Level 0: binary */

Sym_Expr* Sym_add_0(Sym_Expr* left, Sym_Expr* right) {
  return Sym_findOrAddBinary(SYM_PLUS, left, right);
}

Sym_Expr* Sym_subtract_0(Sym_Expr* left, Sym_Expr* right) {
  return Sym_findOrAddBinary(SYM_MINUS, left, right);
}

Sym_Expr* Sym_multiply_0(Sym_Expr* left, Sym_Expr* right) {
  return Sym_findOrAddBinary(SYM_TIMES, left, right);
}

Sym_Expr* Sym_divide_0(Sym_Expr* left, Sym_Expr* right) {
  return Sym_findOrAddBinary(SYM_DIVIDE, left, right);
}

Sym_Expr* Sym_equals_0(Sym_Expr* left, Sym_Expr* right) {
  return Sym_findOrAddBinary(SYM_EQUALS, left, right);
}

Sym_Expr* Sym_nequals_0(Sym_Expr* left, Sym_Expr* right) {
  return Sym_findOrAddBinary(SYM_NEQUALS, left, right);
}

Sym_Expr* Sym_lessThan_0(Sym_Expr* left, Sym_Expr* right) {
  return Sym_findOrAddBinary(SYM_LT, left, right);
}

Sym_Expr* Sym_greaterThan_0(Sym_Expr* left, Sym_Expr* right) {
  return Sym_findOrAddBinary(SYM_GT, left, right);
}

Sym_Expr* Sym_lessThanOrEquals_0(Sym_Expr* left, Sym_Expr* right) {
  return Sym_findOrAddBinary(SYM_LTE, left, right);
}

Sym_Expr* Sym_greaterThanOrEquals_0(Sym_Expr* left, Sym_Expr* right) {
  return Sym_findOrAddBinary(SYM_GTE, left, right);
}

Sym_Expr* Sym_intLessThan_0(Sym_Expr* left, Sym_Expr* right) {
  return Sym_findOrAddBinary(SYM_BLT, left, right);
}

Sym_Expr* Sym_intGreaterThanOrEquals_0(Sym_Expr* left, Sym_Expr* right) {
  return Sym_findOrAddBinary(SYM_BGTE, left, right);
}

Sym_Expr* Sym_conjunct_0(Sym_Expr* left, Sym_Expr* right) {
  return Sym_findOrAddBinary(SYM_AND, left, right);
}


/* Level 0: unary */

Sym_Expr* Sym_negate_0(Sym_Expr* left) {
  return Sym_findOrAddUnary(SYM_NEG, left);
}

Sym_Expr* Sym_sqrt_0(Sym_Expr* left) {
  return Sym_findOrAddUnary(SYM_SQRT, left);
}

Sym_Expr* Sym_abs_0(Sym_Expr* left) {
  return Sym_findOrAddUnary(SYM_ABS, left);
}


/* Level 0: leaf */

Sym_Expr* Sym_intConstant_0(int val) {
  return Sym_findOrAddLeaf(SYM_INT, val);
}

Sym_Expr* Sym_symbolicConstant_0(int val) {
  return Sym_findOrAddLeaf(SYM_SYM_CONST, val);
}


/* Selected Level 1 operations */

Sym_Expr* Sym_add_1(Sym_Expr* left, Sym_Expr* right) {
  if (left == Sym_zero) {
    return right;
  }
  if (right == Sym_zero) {
    return left;
  }
  if (left->operator==SYM_NEG && left->left==right) {
    return Sym_zero;
  }
  if (right->operator==SYM_NEG && right->left==left) {
    return Sym_zero;
  }
  return Sym_add_0(left, right);
}


Sym_Expr* Sym_subtract_1(Sym_Expr* left, Sym_Expr* right) {
  if (right == Sym_zero) {
    return left;
  }
  if (left==right) {
    return Sym_zero;
  }
  return Sym_subtract_0(left,right);
}


Sym_Expr* Sym_multiply_1(Sym_Expr* left, Sym_Expr* right) {
  if (left==Sym_zero || right==Sym_zero) {
    return Sym_zero;
  }
  if (left==Sym_one) {
    return right;
  }
  if (right==Sym_one) {
    return left;
  }
  return Sym_multiply_0(left,right);
}

Sym_Expr* Sym_divide_1(Sym_Expr* left, Sym_Expr* right) {
  Sym_assert(right!=Sym_zero, "divide_1: division by 0");
  if (left==Sym_zero) {
    return Sym_zero;
  }
  if (right==Sym_one) {
    return left;
  }
  if (left==right) {
    return Sym_one;
  }
  return Sym_divide_0(left, right);
}

Sym_Expr* Sym_negate_1(Sym_Expr* left) {
  if (left==Sym_zero) {
    return Sym_zero;
  }
  return Sym_negate_0(left);
}

Sym_Expr* Sym_sqrt_1(Sym_Expr* left) {
  if (left==Sym_zero) {
    return Sym_zero;
  }
  if (left==Sym_one) {
    return Sym_one;
  }
  return Sym_sqrt_0(left);
}

Sym_Expr* Sym_abs_1(Sym_Expr* left) {
  if (left==Sym_zero) {
    return Sym_zero;
  }
  if (left==Sym_one) {
    return Sym_one;
  }
  if (Sym_isInt(left)) {
    int val = Sym_intVal(left);
    
    if (val < 0) val = -val;
    return Sym_intConstant_0(val);
  }
  return Sym_abs_0(left);
}


Sym_Expr* Sym_conjunct_1(Sym_Expr* left, Sym_Expr* right) {
  if (left==Sym_true) {
    return right;
  }
  if (right==Sym_true) {
    return left;
  }
  if (left == Sym_false) {
    return Sym_false;
  }
  if (right == Sym_false) {
    return Sym_false;
  }
  return Sym_conjunct_0(left, right);
}



/* Selected Level 2 operations */


Sym_Expr* Sym_add_2a(Sym_Expr* left, Sym_Expr* right) {
  int j,k;
  Sym_Expr* result;
  Sym_Expr* x = right;

  j=0;
  while (x->operator==SYM_PLUS) {
    Sym_assert(x->right->operator != SYM_PLUS, "bad normal form 2a");
    x = x->left;
    j++;
  }
  result = Sym_add_1(left,x);
  while (j>0) {
    k=1;
    x=right;
    while (k<j) {
      x=x->left;
      k++;
    }
    x=x->right;
    result=Sym_add_1(result,x);
    j--;
  }
  return result;
}  


/* Support functions for add_2b: The "sumlist" normal form : either 0,
 * a, p, or a+p, where p is a non-0 integer constant, and a is a sum
 * of the form ((w+x)+y)+z (etc), where w is not a sum, and
 * furthermore, w<=x<=y<=z (in the global order on expressions).
 * Furthermore, none of the w,x,y,z (etc) should be integer constants.
 * Moreover, there are not two non-integral terms of the form u and
 * -u.  (These cancel to 0.)
 *
 * p is called the integral term and w,x, etc are called the
 * non-integral terms.  The length is the number of non-integral terms
 * (4 in this example).
 */

Sym_SumList* Sym_newSumList() {
  Sym_SumList* result = (Sym_SumList*)malloc(sizeof(Sym_SumList));

  Sym_assert(result!=0, "Sym_newSumList: malloc failed");
  result->length = 0;
  result->intterm = 0;
  return result;
}


void Sym_setList(Sym_SumList* f, int index, Sym_Expr* expr) {
  Sym_assert(index<SYM_MAX_DEPTH, "setList: index >= SYM_MAX_DEPTH");
  Sym_assert(index<f->length, "setList: index >= f->length");
  Sym_assert(expr!=0 , "setList: expr null");
  f->list[index] = expr;
}


Sym_Expr* Sym_getList(Sym_SumList* f, int index) {
  Sym_assert(index<SYM_MAX_DEPTH, "getList: index >= SYM_MAX_DEPTH");
  Sym_assert(index<f->length, "getList: index >= f->length");
  return f->list[index];
}


void Sym_printSumList(Sym_SumList* f) {
  int i;

  Sym_assert(f!=0, "printSumList: null f");
  printf("F[%d,%d", f->length,f->intterm);
  printf(",(");
  for (i=0; i<f->length; i++) {
    if (i>0) {
      printf(",");
    }
    Sym_printExprShort(Sym_getList(f, i));
  }
  printf(")]");
}



/* Given an expression in sumlist normal form, deconstruct to get its
 * (1) list of non-integral terms, and (2) its integral term.  The int
 * length is the length of the list, i.e., the number of non-integral
 * terms.  If there is no integral term, intterm will be set to -1.
 * The list returned gives the non-integral terms in increasing order,
 * i.e., w,x,y,z,..., where w is the 0-th element,x is the 1-th
 * element, etc.
 */
void Sym_exprToSumList(Sym_Expr* expr, Sym_SumList* result) {
  Sym_Expr* expr2= 0; /* the sum of the non-int terms */

  Sym_assert(expr!=0, "exprToSumList: null expr");
  Sym_assert(result!=0, "exprToSumList: null result");
#if DEBUG_CONVERT>0
  printf("exprToSumList: called on e%d\n", expr->id);
#endif
  if (expr->operator==SYM_PLUS) {
    Sym_Expr* lastterm = expr->right;
    
    if (Sym_isInt(lastterm)) {
      result->intterm = Sym_intVal(lastterm);
      expr2 = expr->left;
    }
    else {
      result->intterm = 0;
      expr2 = expr;
    }
  }
  else {
    if (Sym_isInt(expr)) {
      result->intterm = Sym_intVal(expr);
      expr2 = 0;
    }
    else {
      result->intterm = 0;
      expr2 = expr;
    }
  }
#if DEBUG_CONVERT>0
  printf("intterm=%d\n", result->intterm);
  printf("expr2=%d\n", expr2);
  if (expr2) {
    printf("expr2->id=%d\n", expr2->id);
  }
#endif
  if (!expr2) {
    result->length = 0;
  }
  else {
    int j;
    Sym_Expr* tmp;
    Sym_Expr* x = expr2;

    result->length = 0;
    while (x->operator == SYM_PLUS) {
      result->length++;
      Sym_setList(result, result->length-1, x->right);
      x = x->left;
    }
    result->length++;
    Sym_setList(result, result->length-1, x);

    /* reverse the list order */

    for (j=0; j<(result->length)/2; j++) {
      tmp = Sym_getList(result, j);
      Sym_setList(result, j, Sym_getList(result, result->length-j-1));
      Sym_setList(result, result->length-j-1, tmp);
    }
  }
#if DEBUG_CONVERT>0
  printf("Converted \n");
  Sym_printExprShort(expr);
  printf("\n to\n");
  Sym_printSumList(result);
  printf("\n");
#endif
}


void Sym_cancelTerms(Sym_SumList *f1, Sym_SumList *f2) {
  int i = 0;

#if DEBUG_CANCEL>0
  printf("Cancel ");
  Sym_printSumList(f1);
  printf(" and ");
  Sym_printSumList(f2);
  printf(" -> ");
#endif
  while (i < f1->length) {
    Sym_Expr* expr1 = Sym_getList(f1, i);
    int cancel = 0;
    int j = 0;

    while (j < f2->length) {
      Sym_Expr* expr2 = Sym_getList(f2, j);

      if ((expr1->operator==SYM_NEG && expr1->left==expr2) ||
	  (expr2->operator==SYM_NEG && expr2->left==expr1)) {
	int k;

	cancel = 1;
	for (k=j; k<f2->length-1; k++) {
	  Sym_setList(f2, k, Sym_getList(f2, k+1));
	}
	f2->length = f2->length - 1;
	break;
      }
      j++;
    }
    if (cancel) {
      int l;

      for (l=i; l<f1->length-1; l++) {
	Sym_setList(f1, l, Sym_getList(f1, l+1));
      }
      f1->length = f1->length - 1;
    }
    else {
      i++;
    }
  }
#if DEBUG_CANCEL>0
  Sym_printSumList(f1);
  printf(" and ");
  Sym_printSumList(f2);
  printf("\n");
#endif    
}


void Sym_addSumList(Sym_SumList* f1, Sym_SumList* f2, Sym_SumList* f3) {
  int i,j,k;

  Sym_cancelTerms(f1, f2);
  f3->length = f1->length + f2->length;
  i = j = k = 0;
  while (i<f1->length && j<f2->length) {
    Sym_Expr* tmp1 = Sym_getList(f1, i);
    Sym_Expr* tmp2 = Sym_getList(f2, j);

    if (tmp1->id <= tmp2->id) {
      Sym_setList(f3, k, Sym_getList(f1, i));
      i++;
    }
    else {
      Sym_setList(f3, k, Sym_getList(f2, j));
      j++;
    }
    k++;
  }
  while (i<f1->length) {
    Sym_setList(f3, k, Sym_getList(f1, i));
    i++;
    k++;
  }
  while (j<f2->length) {
    Sym_setList(f3, k, Sym_getList(f2, j));
    j++;
    k++;
  }
  f3->intterm=f1->intterm+f2->intterm;
}


Sym_Expr* Sym_SumListToExpr(Sym_SumList *f) {
  int i;
  Sym_Expr* expr = Sym_zero;

  for (i=0; i<f->length; i++) {
    expr = Sym_add_1(expr, Sym_getList(f, i));
  }
  if (f->intterm != 0) {
    expr = Sym_add_1(expr, Sym_intConstant_0(f->intterm));
  }
  return expr;
}


Sym_Expr* Sym_add_2b(Sym_Expr* left, Sym_Expr* right) {
  Sym_Expr* expr;

  Sym_assert(left!=0, "add_2b: null left");
  Sym_assert(right!=0, "add_2b: null right");
#if DEBUG_ADD3>0
  printf("Calling add_3 on e%d and e%d.\n\n", left->id, right->id);
  Sym_printExpr(left);
  printf("\n\n");
  Sym_printExpr(right);
  printf("\n\n");
#endif
  Sym_exprToSumList(left,Sym_sumlist1);
  Sym_exprToSumList(right,Sym_sumlist2);
  Sym_addSumList(Sym_sumlist1, Sym_sumlist2, Sym_sumlist3);
  expr = Sym_SumListToExpr(Sym_sumlist3);
#if DEBUG_ADD3>0
  printf("add_3 returning e%d.\n", expr->id);
  printf("\n\n");
  Sym_printExpr(expr);
  printf("\n\n");
#endif
  return expr;
}


/* Generic arithmetic functions: these pick the best version
 * available, within the bound given by Sym_equivalenceType.
 */

Sym_Expr* Sym_add(Sym_Expr* left, Sym_Expr* right) {
  if (Sym_equivalenceType==0) return Sym_add_0(left,right);
  if (Sym_equivalenceType==1) return Sym_add_1(left,right);
  return Sym_add_2b(left,right);
}

/* For integer addition, assoc/comm are allowed when
 * checking IEEE equivalence */
Sym_Expr* Sym_add_int(Sym_Expr* left, Sym_Expr* right) {
  if (Sym_equivalenceType==0) return Sym_add_0(left,right);
  return Sym_add_2b(left,right);
}

Sym_Expr* Sym_subtract(Sym_Expr* left, Sym_Expr* right) {
  if (Sym_equivalenceType==0) return Sym_subtract_0(left,right);
  return Sym_subtract_1(left,right);
}

Sym_Expr* Sym_multiply(Sym_Expr* left, Sym_Expr* right) {
  if (Sym_equivalenceType==0) return Sym_multiply_0(left,right);
  return Sym_multiply_1(left,right);
}

Sym_Expr* Sym_divide(Sym_Expr* left, Sym_Expr* right) {
  if (Sym_equivalenceType==0) return Sym_divide_0(left,right);
  return Sym_divide_1(left,right);
}

Sym_Expr* Sym_equals(Sym_Expr* left, Sym_Expr* right) {
  return Sym_equals_0(left,right);
}

Sym_Expr* Sym_nequals(Sym_Expr* left, Sym_Expr* right) {
  return Sym_nequals_0(left,right);
}

Sym_Expr* Sym_lessThan(Sym_Expr* left, Sym_Expr* right) {
  return Sym_lessThan_0(left,right);
}

Sym_Expr* Sym_greaterThan(Sym_Expr* left, Sym_Expr* right) {
  return Sym_greaterThan_0(left,right);
}

Sym_Expr* Sym_lessThanOrEquals(Sym_Expr* left, Sym_Expr* right) {
  return Sym_lessThanOrEquals_0(left,right);
}

Sym_Expr* Sym_greaterThanOrEquals(Sym_Expr* left, Sym_Expr* right) {
  return Sym_greaterThanOrEquals_0(left,right);
}

Sym_Expr* Sym_intLessThan(Sym_Expr* left, Sym_Expr* right) {
  return Sym_intLessThan_0(left,right);
}

Sym_Expr* Sym_intGreaterThanOrEquals(Sym_Expr* left, Sym_Expr* right) {
  return Sym_intGreaterThanOrEquals_0(left,right);
}

Sym_Expr* Sym_conjunct(Sym_Expr* left, Sym_Expr* right) {
  return Sym_conjunct_1(left,right);
}

Sym_Expr* Sym_negate(Sym_Expr* left) {
  if (Sym_equivalenceType==0) return Sym_negate_0(left);
  return Sym_negate_1(left);
}

Sym_Expr* Sym_sqrt(Sym_Expr* left) {
  if (Sym_equivalenceType==0) return Sym_sqrt_0(left);
  return Sym_sqrt_1(left);
}

Sym_Expr* Sym_abs(Sym_Expr* left) {
  if (Sym_equivalenceType==0) return Sym_abs_0(left);
  return Sym_abs_1(left);
}

Sym_Expr* Sym_intConstant(int val) {
  return Sym_intConstant_0(val);
}

Sym_Expr* Sym_symbolicConstant(int val) {
  return Sym_symbolicConstant_0(val);
}


/* Looking up facts in path conditions... */

/* Note: these assume the path condition pc is an expression of the
 * following form (AND (AND (AND e1) e2) e3), etc.  With a little bit
 * of work, we could make this more general, but it is fine for now.
 *
 * This will return 1 if pc implies that expr1==expr2, 0 if pc implies
 * that expr1!=expr2, and 2 if it can't decide.
 */
int Sym_isEqualTo(Sym_Expr* pc, Sym_Expr* expr1, Sym_Expr* expr2) {
  int done, result;
  Sym_Expr* expr;

  if (expr1 == expr2) {
    result = 1;
  }
  else {
    result = 2; /* "I don't know" */
    done = 0;
    while (!done) {
      if (pc->operator == SYM_AND) {
	expr = pc->right;
	pc = pc->left;
      }
      else {
	expr = pc;
	done = 1;
      }
      if (expr == Sym_true) {
      }
      else if (expr == Sym_false) {
	result = 1;
	break;
      }
      else if (expr->operator==SYM_EQUALS &&
	       ((expr->left==expr1 && expr->right==expr2) ||
		(expr->left==expr2 && expr->right==expr1))) {
	result = 1;
	break;
      }
      else if ((expr->operator==SYM_NEQUALS ||
		expr->operator==SYM_LT ||
		expr->operator==SYM_GT) &&
	       ((expr->left==expr1 && expr->right==expr2) ||
		(expr->left==expr2 && expr->right==expr1))) {
	result = 0;
	break;
      }
    }
  }
#if DEBUG_ISEQ
  printf("SYM_isEqualTo(e%d,e%d): %d\n", expr1->id, expr2->id, result);
#endif
  return result;
}



/* See notes for isEqualTo above. */
int Sym_isLessThan(Sym_Expr* pc, Sym_Expr* expr1, Sym_Expr* expr2) {
  int done, result;
  Sym_Expr* expr;

  if (Sym_isInt(expr1) && Sym_isInt(expr2)) {
    if (Sym_intVal(expr1) < Sym_intVal(expr2)) {
      result = 1;
    }
    else {
      result = 0;
    }
  }
  else {
    result = 2; /* "I don't know" */
    done = 0;
    while (!done) {
      if (pc->operator==SYM_AND) {
	expr = pc->right;
	pc = pc->left;
      }
      else {
	expr = pc;
	done = 1;
      }
      if (expr == Sym_true) {
      }
      else if (expr == Sym_false) {
	result = 1;
	break;
      }
      else if (expr->operator==SYM_LT && expr->left==expr1 &&
	       expr->right==expr2) {
	result = 1;
	break;
      }
      else if ((expr->operator==SYM_GTE || expr->operator==SYM_GT ||
		expr->operator==SYM_EQUALS) &&
	       expr->left==expr1 && expr->right==expr2) {
	result = 0;
	break;
      }
    }
  }
#if DEBUG_LT
  /*  printf("isLessThan: returning %d\n", result);*/
  if ((result != 0) && (result != 1)) {
    printf(" isLessThan: unable to determine %d < %d\n", expr1->id, expr2->id);
    printf("e%d: ", expr1->id);
    Sym_printExprShort(expr1);
    printf("\n");
    Sym_printExpr(expr1);
    printf("\n\n");
    printf("e%d: ", expr2->id);
    Sym_printExprShort(expr2);
    printf("\n");
    Sym_printExpr(expr2);
    printf("\n\nSym Table:\n\n");
    Sym_printSymTableShort();
    printf("\n\n");
  }
#endif
  return result;
}


/* Note: this assumes the path condition pc is an expression of the
 * following form (AND (AND (AND e1) e2) e3), etc.  With a little bit
 * of work, we could make this more general, but it is fine for now.
 * It returns the number of terms, not including any which are just
 * "true". */
int Sym_pathConditionLength(Sym_Expr* pc) {
  int done, count;
  Sym_Expr* expr;

  count = 0;
  done = 0;
  while (!done) {
    if (pc->operator==SYM_AND) {
      expr = pc->right;
      pc = pc->left;
    }
    else {
      expr = pc;
      done = 1;
    }
    if (expr == Sym_true) {
    }
    else {
      count++;
    }
  }
  return count;
}



int Sym_maxChainLength() {
  int i,j,max;
  Sym_Expr* expr;
  
  max=0;
  for (i=0; i<SYM_HASHTABLE_LEN; i++) {
    expr = Sym_hashtable[i];
    j=0;
    while (expr) {
      j++;
      expr=expr->next;
    }
    max=MAX(max,j);
  }
  return max;
}
    


/* Public functions (begin with SYM) */

int SYM_init() {
  Sym_sumlist1=Sym_newSumList();
  Sym_sumlist2=Sym_newSumList();
  Sym_sumlist3=Sym_newSumList();
  Sym_zero=Sym_addLeaf(SYM_INT, 0);
  Sym_one=Sym_addLeaf(SYM_INT, 1);
  Sym_false=Sym_zero;
  Sym_true=Sym_one;
  return 2;
}

int SYM_tableLength() {
  return Sym_nExpr;
}

int SYM_maxChainLength() {
  return Sym_maxChainLength();
}

int SYM_printExpr(int id) {
  return Sym_printExpr(Sym_getExpr(id));
}

void SYM_printExprShort(int id) {
  Sym_printExprShort(Sym_getExpr(id));
}

int SYM_exprLength(int id) {
  return Sym_exprLength(Sym_getExpr(id));
}

void SYM_printSymTable() {
  Sym_printSymTable();
}


void SYM_printSymTableShort() {
  Sym_printSymTableShort();
}

void SYM_printPathCondition(int pc) {
  Sym_printPathCondition(Sym_getExpr(pc));
}

void SYM_setEquivalenceType(int val) {
  Sym_equivalenceType = val;
}

int SYM_getEquivalenceType() {
  return Sym_equivalenceType;
}

int SYM_isInt(int id) {
  return Sym_isInt(Sym_getExpr(id));
}

int SYM_intVal(int id) {
  return Sym_intVal(Sym_getExpr(id));
}


int SYM_add_0(int operand1, int operand2) {
  return Sym_add_0(Sym_getExpr(operand1),
		   Sym_getExpr(operand2))->id;
}

int SYM_subtract_0(int operand1, int operand2) {
  return Sym_subtract_0(Sym_getExpr(operand1),
			Sym_getExpr(operand2))->id;
}
int SYM_multiply_0(int operand1, int operand2) {
  return Sym_multiply_0(Sym_getExpr(operand1),
			Sym_getExpr(operand2))->id;
}
int SYM_divide_0(int operand1, int operand2) {
  return Sym_divide_0(Sym_getExpr(operand1),
		      Sym_getExpr(operand2))->id;
}
int SYM_equals_0(int operand1, int operand2) {
  return Sym_equals_0(Sym_getExpr(operand1),
		      Sym_getExpr(operand2))->id;
}
int SYM_nequals_0(int operand1, int operand2) {
  return Sym_nequals_0(Sym_getExpr(operand1),
		       Sym_getExpr(operand2))->id;
}
int SYM_lessThan_0(int operand1, int operand2) {
  return Sym_lessThan_0(Sym_getExpr(operand1),
			Sym_getExpr(operand2))->id;
}
int SYM_greaterThan_0(int operand1, int operand2) {
  return Sym_greaterThan_0(Sym_getExpr(operand1),
			   Sym_getExpr(operand2))->id;
}

int SYM_lessThanOrEquals_0(int operand1, int operand2) {
  return Sym_lessThanOrEquals_0(Sym_getExpr(operand1),
				Sym_getExpr(operand2))->id;
}

int SYM_greaterThanOrEquals_0(int operand1, int operand2) {
  return Sym_greaterThanOrEquals_0(Sym_getExpr(operand1),
				   Sym_getExpr(operand2))->id;
}
int SYM_intLessThan_0(int operand1, int operand2) {
  return Sym_intLessThan_0(Sym_getExpr(operand1),
			   Sym_getExpr(operand2))->id;
}

int SYM_intGreaterThanOrEquals_0(int operand1, int operand2) {
  return Sym_intGreaterThanOrEquals_0(Sym_getExpr(operand1),
				      Sym_getExpr(operand2))->id;
}

int SYM_conjunct_0(int operand1, int operand2) {
  return Sym_conjunct_0(Sym_getExpr(operand1),
			Sym_getExpr(operand2))->id;
}

int SYM_negate_0(int operand) {
  return Sym_negate_0(Sym_getExpr(operand))->id;
}

int SYM_sqrt_0(int operand) {
  return Sym_sqrt_0(Sym_getExpr(operand))->id;
}

int SYM_abs_0(int operand) {
  return Sym_abs_0(Sym_getExpr(operand))->id;
}

int SYM_intConstant_0(int operand) {
  return Sym_intConstant_0(operand)->id;
}

int SYM_symbolicConstant_0(int operand) {
  return Sym_symbolicConstant_0(operand)->id;
}

int SYM_add_1(int operand1, int operand2) {
  return Sym_add_1(Sym_getExpr(operand1), Sym_getExpr(operand2))->id;
}

int SYM_subtract_1(int operand1, int operand2) {
  return Sym_subtract_1(Sym_getExpr(operand1),Sym_getExpr(operand2))->id;
}

int SYM_multiply_1(int operand1, int operand2) {
  return Sym_multiply_1(Sym_getExpr(operand1),Sym_getExpr(operand2))->id;
}

int SYM_divide_1(int operand1, int operand2) {
  return Sym_divide_1(Sym_getExpr(operand1),Sym_getExpr(operand2))->id;
}

int SYM_negate_1(int operand) {
  return Sym_negate_1(Sym_getExpr(operand))->id;
}

int SYM_sqrt_1(int operand) {
  return Sym_sqrt_1(Sym_getExpr(operand))->id;
}

int SYM_abs_1(int operand) {
  return Sym_abs_1(Sym_getExpr(operand))->id;
}

int SYM_conjunct_1(int operand1, int operand2) {
  return Sym_conjunct_1(Sym_getExpr(operand1),Sym_getExpr(operand2))->id;
}




int SYM_add(int operand1, int operand2) {
  return Sym_add(Sym_getExpr(operand1),
		 Sym_getExpr(operand2))->id;
}


/* For integer addition, assoc/comm are allowed when
 * checking IEEE equivalence */
int SYM_add_int(int operand1, int operand2) {
  return Sym_add_int(Sym_getExpr(operand1),
		     Sym_getExpr(operand2))->id;
}

int SYM_subtract(int operand1, int operand2) {
  return Sym_subtract(Sym_getExpr(operand1),
		      Sym_getExpr(operand2))->id;
}
int SYM_multiply(int operand1, int operand2) {
  return Sym_multiply(Sym_getExpr(operand1),
		      Sym_getExpr(operand2))->id;
}
int SYM_divide(int operand1, int operand2) {
  return Sym_divide(Sym_getExpr(operand1),
		    Sym_getExpr(operand2))->id;
}
int SYM_equals(int operand1, int operand2) {
  return Sym_equals(Sym_getExpr(operand1),
		    Sym_getExpr(operand2))->id;
}
int SYM_nequals(int operand1, int operand2) {
  return Sym_nequals(Sym_getExpr(operand1),
		     Sym_getExpr(operand2))->id;
}
int SYM_lessThan(int operand1, int operand2) {
  return Sym_lessThan(Sym_getExpr(operand1),
		      Sym_getExpr(operand2))->id;
}
int SYM_greaterThan(int operand1, int operand2) {
  return Sym_greaterThan(Sym_getExpr(operand1),
			 Sym_getExpr(operand2))->id;
}

int SYM_lessThanOrEquals(int operand1, int operand2) {
  return Sym_lessThanOrEquals(Sym_getExpr(operand1),
			      Sym_getExpr(operand2))->id;
}

int SYM_greaterThanOrEquals(int operand1, int operand2) {
  return Sym_greaterThanOrEquals(Sym_getExpr(operand1),
				 Sym_getExpr(operand2))->id;
}
int SYM_intLessThan(int operand1, int operand2) {
  return Sym_intLessThan(Sym_getExpr(operand1),
			 Sym_getExpr(operand2))->id;
}

int SYM_intGreaterThanOrEquals(int operand1, int operand2) {
  return Sym_intGreaterThanOrEquals(Sym_getExpr(operand1),
				    Sym_getExpr(operand2))->id;
}

int SYM_conjunct(int operand1, int operand2) {
  return Sym_conjunct(Sym_getExpr(operand1),
		      Sym_getExpr(operand2))->id;
}

int SYM_negate(int operand) {
  return Sym_negate(Sym_getExpr(operand))->id;
}

int SYM_sqrt(int operand) {
  return Sym_sqrt(Sym_getExpr(operand))->id;
}

int SYM_abs(int operand) {
  return Sym_abs(Sym_getExpr(operand))->id;
}

int SYM_intConstant(int operand) {
  return Sym_intConstant(operand)->id;
}

int SYM_symbolicConstant(int operand) {
  return Sym_symbolicConstant(operand)->id;
}


int SYM_isEqualTo(int pc, int operand1, int operand2) {
  return Sym_isEqualTo(Sym_getExpr(pc), Sym_getExpr(operand1),
		       Sym_getExpr(operand2));
}

int SYM_isLessThan(int pc, int operand1, int operand2) {
  return Sym_isLessThan(Sym_getExpr(pc), Sym_getExpr(operand1),
			Sym_getExpr(operand2));
}

int SYM_add_2a(int operand1, int operand2) {
  return Sym_add_2a(Sym_getExpr(operand1),Sym_getExpr(operand2))->id;
}

int SYM_add_2b(int operand1, int operand2) {
  return Sym_add_2b(Sym_getExpr(operand1),Sym_getExpr(operand2))->id;
}


int SYM_pathConditionLength(int pc) {
  return Sym_pathConditionLength(Sym_getExpr(pc));
}

