/*
 * Copyright (c) 2004, University of Massachusetts
 * All Rights Reserved.
 * 
 * Name:       matmat.c
 * Date:       08 Jul 2004
 * Revised:    09 Jul 2004
 * Author:     Stephen Siegel <siegel@cs.umass.edu>
 * Maintainer: Stephen Siegel
 * Reader:
 *
 * Compile:    mpicc matmat.c
 * Run:        mpirun -np N a.out
 *
 *   N : number of slaves + 1
 *
 *   Note: number of slaves must be less than or equal to 
 *   arows
 *
 * This code is adapted from an example in the book Using MPI, by
 * Gropp, Lusk, and Skjellum.  It takes the produce of two matrices,
 * using a master-slave appraoch to achieve automatic load-balancing.
 */

#include<stdio.h>
#include "mpi.h"

#define MAX_AROWS 20
#define MAX_ACOLS 100
#define MAX_BCOLS 20
#define MAX_A_ENTRIES 2000
#define MAX_B_ENTRIES 2000
#define MAX_C_ENTRIES 400

int main(int argc,char *argv[]) {

  double a[MAX_A_ENTRIES], b[MAX_B_ENTRIES], c[MAX_C_ENTRIES];
  double buffer[MAX_ACOLS], ans[MAX_ACOLS];

  int myid, master, numprocs, ierr;
  int i, j, numsent, sender;
  int anstype, row, arows, acols, brows, bcols, crows, ccols;
  MPI_Status status;

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
  MPI_Comm_rank(MPI_COMM_WORLD, &myid);
  
  master = 0;
  arows  = 20;
  acols  = 4;
  brows  = 4;
  bcols  = 3;
  crows  = arows;
  ccols  = bcols;

  if (myid == master) {
    for (i = 0; i < acols; i++) {
      for (j = 0; j < arows; j++) {
	a[j*acols+i] = i+1;
      }
    }
    for (i = 0; i < bcols; i++) {
      for (j = 0; j < brows; j++) {
	b[j*bcols+i] = i+1;
      }
    }
    for (i = 0; i < arows; i++) {
      for (j = 0; j < acols; j++) {
	printf("a[%d,%d]=%f\n", i, j, a[i*acols+j]);
      }
    }
    printf("\n");
    for (i = 0; i < brows; i++) {
      for (j = 0; j < bcols; j++) {
	printf("b[%d,%d]=%f\n", i, j, b[i*bcols+j]);
      }
    }
    printf("\n");
    numsent = 0;
    MPI_Bcast(b, brows*bcols, MPI_DOUBLE, master, MPI_COMM_WORLD);
    for (i = 0; i < numprocs-1; i++) {
      for (j = 0; j < acols; j++) {
	buffer[j] = a[i*acols+j];
      }
      MPI_Send(buffer, acols, MPI_DOUBLE, i+1, i+1, MPI_COMM_WORLD);
      numsent++;
    }
    for (i = 0; i < crows; i++) {
      MPI_Recv(ans, ccols, MPI_DOUBLE, MPI_ANY_SOURCE, MPI_ANY_TAG, 
	       MPI_COMM_WORLD, &status);
      sender = status.MPI_SOURCE;
      anstype = status.MPI_TAG - 1;
      for (j = 0; j < ccols; j++) {
	c[anstype*ccols+j] = ans[j];
      }
      if (numsent < arows) {
	for (j = 0; j < acols; j++) {
	  buffer[j] = a[numsent*acols+j];
	}
	MPI_Send(buffer, acols, MPI_DOUBLE, sender, numsent+1, MPI_COMM_WORLD);
	numsent++;
      }
      else {
	MPI_Send(buffer, 1, MPI_DOUBLE, sender, 0, MPI_COMM_WORLD);
      }
    }
    for (i = 0; i < crows; i++) {
      for (j = 0; j < ccols; j++) {
	printf("c[%d,%d]=%f\n", i, j, c[i*ccols+j]);
      }
    }
  }
  else {
    MPI_Bcast(b, brows*bcols, MPI_DOUBLE, master, MPI_COMM_WORLD);
    while (1) {
      MPI_Recv(buffer, acols, MPI_DOUBLE, master, MPI_ANY_TAG,
	       MPI_COMM_WORLD, &status);
      if (status.MPI_TAG == 0) break;
      row = status.MPI_TAG - 1;
      for (i = 0; i < bcols; i++) {
	ans[i] = 0.0;
	for (j = 0; j < acols; j++) {
	  ans[i] += buffer[j]*b[j*bcols+i];
	}
      }
      MPI_Send(ans, bcols, MPI_DOUBLE, master, row+1, MPI_COMM_WORLD);
    }
  }
  MPI_Finalize();
}
